//Various

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <math.h>

//Mygraph
#include "mygraph.hpp"
#include "path.hpp"
#include "components.hpp"

//Mersenne
#include <Mersenne/MersenneTwister.h>

typedef vector<int> VertexMapI;
typedef vector<double> VertexMapD;
typedef vector<vector<int> > IntMtx;
typedef vector<pair<int,int> > pairvectorii;

MyGraph make_random_net(int n, int m);
void update_vertices(const int &time, VertexMapI &infected, const VertexMapD &th, IntMtx &fads, const MyGraph &G);


MTRand *R;

const int ORIGIN=0;
const int INIT=1;
const int DEAD=2;
const int NCURR=3;
const int NMAX=4;
const int TMAX=5;
const int DEGREE=6;
const int EXC=7;

int main(int argc, char* argv[])
{

  int n = atoi(argv[1]);
  int m = atoi(argv[2]);
  double TH = atof(argv[3]);
  int NT = 12800;
  R = new MTRand();


  for(int rlz=0;rlz<250;rlz++){
    
    MyGraph Graph = make_random_net(n,m);
    MyGraph GC = components(Graph);
    int nc = num_vertices(GC);
    vector<int> exc(nc);

    for(int i=0;i<nc;i++)
      exc[i] = excentricity(i,GC);
    
    IntMtx fads;    
    fads.reserve(NT);
    vector<int> fad(8,0);
    VertexMapI infected(nc,0);
    VertexMapD th(nc,TH);
    int nfads = 1;
    fad[ORIGIN] = R->randInt(nc-1);
    fad[INIT] = 0;
    fad[DEAD] = -1;
    fad[NCURR] = nc;
    fad[NMAX] = nc;
    fad[TMAX] = 0;
    fad[DEGREE] = degree(fad[ORIGIN],GC);
    fad[EXC] = exc[fad[ORIGIN]];
    infected[fad[ORIGIN]] = 0;
    
    fads.push_back(fad);
    
    for(int t=1 ; t < NT ; t++){
      
      update_vertices(t,infected,th,fads,GC);
      
      if(R->randInt(6399) < nc){
	nfads++;
	fad[ORIGIN] = R->randInt(nc-1);
	fad[INIT] = t;
	fad[DEAD] = -1;
	fad[NCURR] = 1;
	fad[NMAX] = 1;
	fad[TMAX] = t;
	fad[DEGREE] = degree(fad[ORIGIN],GC);
	fad[EXC] = exc[fad[ORIGIN]];
	infected[fad[ORIGIN]] = nfads - 1;
	fads.push_back(fad);
      }
      
    }
    
    
    char filename[132];
    
    sprintf(filename,"data/%d/data_%d_%d_%g.dat",n,n,m,TH);
    
    ofstream file(filename,ios::app);
    
    for(int i=1; i < fads.size() ; i++){
      if(fads[i][DEAD] != -1)
	file << (fads[i][DEAD] - fads[i][INIT]) << " " << (fads[i][TMAX] - fads[i][INIT]) << " " << fads[i][NMAX]
	     << " " << fads[i][DEGREE] << " " << fads[i][EXC] << endl;
    }

  }
    
    return(1);

}


MyGraph make_random_net(int n, int m){

  Vertex n1,n2;

  MyGraph G(n);

  while(num_edges(G) < m){
    
    n1 = R->randInt(n-1);
    
    do{
      n2 = R->randInt(n-1);
    }  while((n1 == n2) || edge(n1,n2,G).second || edge(n2,n1,G).second);
    
    add_edge(n1,n2,G);

  }

  return(G);

}

void update_vertices(const int &time, VertexMapI &infected, const VertexMapD &th, IntMtx &fads, const MyGraph &G){

  int n = num_vertices(G);

  vector<pair<double,int> > max_score(n, pair<double,int>(-1.0 , -1));

  for(int i=0; i < n; i++){

    list<int> inf;

    Vertex v = i;

    OutEdgeIt oi = out_edges(v,G).first;
    OutEdgeIt oe = out_edges(v,G).second;

    while(oi != oe){
      inf.push_back(infected[target(*oi,G)]); 
      oi++;
    }
   
    InEdgeIt ii = in_edges(v,G).first;
    InEdgeIt ie = in_edges(v,G).second;

    while(ii != ie){
      inf.push_back(infected[source(*ii,G)]);
      ii++;
    }

    vector<pair<int,int> > nv(inf.size(),pair<int,int>(-1 , 0));
    int numfads = 0;
    list<int>::iterator li,lb,le;
 
    lb = inf.begin();
    le = inf.end();

    for(li = lb; li != le; li++){

      int j = 0;
      bool inserted = false;

      while(!inserted){
	
	if(*li == nv[j].first){
	  nv[j].second++;
	  inserted = true;
	}
	if(nv[j].first == -1){
	  nv[j].first = *li;
	  nv[j].second = 1;
	  numfads++;
	  inserted = true;
	}
	
	j++;

      }
    }
    
    double sfvt;
    double k = ((2.0*num_edges(G))/(1.0*num_vertices(G)*degree(v,G)))/(1.0*time - 1.0*fads[infected[v]][INIT]);

    for(int j=0 ; j < numfads ; j++){ // Calculate score
      sfvt = k * nv[j].second*(fads[nv[j].first][INIT] - fads[infected[v]][INIT]);
      
      if(sfvt > max_score[v].first){
	max_score[v].first = sfvt;
	max_score[v].second = nv[j].first;
      }
    }
    
  }
  
  for(int i=0; i < n; i++){ // Update if max_score > threshold

    Vertex v = i;

    if(max_score[v].first > th[v]){
      
      fads[infected[v]][NCURR]--;

      if(fads[infected[v]][NCURR] == 0)
	fads[infected[v]][DEAD] = time;

      infected[v] = max_score[v].second;

      fads[infected[v]][NCURR]++;

      if(fads[infected[v]][NCURR] > fads[infected[v]][NMAX]){
	fads[infected[v]][NMAX] = fads[infected[v]][NCURR];
	fads[infected[v]][TMAX] = time;
      }
    }
  }  
}

